/*
 * Copyright (C) 2008 Till Harbaum <till@harbaum.org>.
 *
 * This file is part of OSM2Go.
 *
 * OSM2Go is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OSM2Go is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OSM2Go.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "appdata.h"

#include <gdk/gdkkeysyms.h>

static void map_way_color(potlatch_conf_t *potlatch, way_t *way) {
  way->color = way->fill_color = MAP_COLOR_NONE;
  way->draw_flags = 0;

  potlatch_color_t *plcol = potlatch->colors;

  /* a highway/waterway/railway is never a polygon */
  if(!osm_way_is_way(way)) {
  
    /* check if this way is a polygon */
    node_chain_t *node_chain = way->node_chain;
    while(node_chain && node_chain->next) node_chain = node_chain->next;
    if(way->node_chain && node_chain && 
       way->node_chain->node == node_chain->node) {
      way->draw_flags |= OSM_DRAW_FLAGS_POLYGON;
      
      /* default fill color is light grey */
     way->fill_color = DEF_FILL_COLOR;
    }
  }
  
  while(plcol) {
    if(osm_way_has_key_or_value(way, plcol->tag)) {

      if(plcol->color != POTLATCH_COLOR_NONE) 
	way->color = (plcol->color << 8) | 0xff;

      if(plcol->stroke) 
	way->draw_flags |= OSM_DRAW_FLAGS_STROKE;
      
      if(plcol->fill_color != POTLATCH_COLOR_NONE)
	way->fill_color = (plcol->fill_color << 8) | FILL_TRANSP;
    }
    plcol = plcol->next;
  }
  
  /* use dark grey/no stroke/not filled for everything unknown */
  if(way->color == MAP_COLOR_NONE)
    way->color = DARK_GREY;
}

static void map_init_colors(potlatch_conf_t *potlatch, osm_t *osm) {
  way_t *way = osm->way;

  printf("preparing colors\n");

  /* colorize ways */
  while(way) {
    map_way_color(potlatch, way);
    way = way->next;
  }
}

static void map_statusbar(map_t *map, map_item_t *map_item) {
  char *item_str = NULL;
  item_id_t id = ID_ILLEGAL;
  tag_t *tag = NULL;
  char *str = NULL;
  
  switch(map_item->type) {
  case MAP_TYPE_NODE:
    item_str = "Node";
    id = map_item->node->id;
    tag = map_item->node->tag;
    break;

  case MAP_TYPE_WAY:
    item_str = "Way";
    id = map_item->way->id;
    tag = map_item->way->tag;
    break;

  default:
    break;
  }

  if(id == ID_ILLEGAL) 
    str = g_strdup_printf(_("Unknown item"));
  else {
    str = g_strdup_printf("%s #%ld", item_str, id);

    /* add some tags ... */
    while(tag) {
      /* we don't have much space, so ignore created_by tag */
      if(!osm_is_creator_tag(tag)) {
	char *old = str;
	str = g_strdup_printf("%s, %s=%s", old, tag->key, tag->value);
	g_free(old);
      }
      tag = tag->next;
    }
  }
  
  statusbar_set(map->appdata, str);
  g_free(str);
}

static void map_outside_error(appdata_t *appdata) {
  errorf(GTK_WIDGET(appdata->window), 
	 _("Items must not be placed outside the working area!"));
}

static void map_item_chain_destroy(map_item_chain_t **chainP) {
  map_item_chain_t *chain = *chainP;
  while(chain) {
    map_item_chain_t *next = chain->next;
    canvas_item_destroy(chain->map_item->item);
    chain = next;
  }

  /* way until gtks event handling has actually destroyed this item */
  printf("waiting for item destruction ");
  while(gtk_events_pending() || *chainP) {
    putchar('.');
    gtk_main_iteration();
  }
  printf(" ok\n");

  /* the callback routine connected to this item should have been */
  /* called by now and it should have set the chain to NULL */
  //  g_assert(!*chainP);
}

static void map_node_select(appdata_t *appdata, node_t *node) {
  map_t *map = appdata->map;
  map_item_t *map_item = &map->selected;
  
  g_assert(!map->highlight);

  map_item->type      = MAP_TYPE_NODE;
  map_item->node      = node;
  map_item->highlight = FALSE;

  /* node may not have any visible representation at all */
  if(node->map_item_chain) 
    map_item->item = node->map_item_chain->map_item->item;
  else
    map_item->item = NULL;

  map_statusbar(map, map_item);
  icon_bar_map_item_selected(appdata, map_item, TRUE);

  /* highlight node */
  gint x = map_item->node->lpos.x, y = map_item->node->lpos.y;

  /* create a copy of this map item and mark it as being a highlight */
  map_item_t *new_map_item = g_new0(map_item_t, 1);
  memcpy(new_map_item, map_item, sizeof(map_item_t));
  new_map_item->highlight = TRUE;
  map_hl_circle_new(map, CANVAS_GROUP_NODES_HL, new_map_item, 
		    x, y, 2*RADIUS, HIGHLIGHT_COLOR);

  if(!map_item->item) {
    /* and draw a fake node */
    new_map_item = g_new0(map_item_t, 1);
    memcpy(new_map_item, map_item, sizeof(map_item_t));
    new_map_item->highlight = TRUE;
    map_hl_circle_new(map, CANVAS_GROUP_NODES_HL, new_map_item, 
		      x, y, RADIUS, HIGHLIGHT2_COLOR);
  }
}

static void map_way_select(appdata_t *appdata, way_t *way) {
  map_t *map = appdata->map;
  map_item_t *map_item = &map->selected;
  
  g_assert(!map->highlight);

  map_item->type      = MAP_TYPE_WAY;
  map_item->way       = way;
  map_item->highlight = FALSE;
  map_item->item      = way->map_item_chain->map_item->item;

  map_statusbar(map, map_item);
  icon_bar_map_item_selected(appdata, map_item, TRUE);

  node_chain_t *node_chain = map_item->way->node_chain;
  while(node_chain) {
    map_item_t item;
    item.type = MAP_TYPE_NODE;
    item.node = node_chain->node;

    if(!map_hl_item_is_highlighted(map, &item)) {

      /* create a new map item for every node */
      map_item_t *new_map_item = g_new0(map_item_t, 1);
      new_map_item->type = MAP_TYPE_NODE;
      new_map_item->node = node_chain->node;
      new_map_item->highlight = TRUE;
    
      gint x = node_chain->node->lpos.x;
      gint y = node_chain->node->lpos.y;

      map_hl_circle_new(map, CANVAS_GROUP_NODES_HL, new_map_item, 
			x, y, RADIUS, HIGHLIGHT2_COLOR);
    }

    node_chain = node_chain->next;
  }

  /* a way needs at least 2 points to be drawn */
  guint nodes = osm_way_number_of_nodes(way);
  if(nodes > 1) {
    
    /* allocate space for nodes */
    canvas_points_t *points = canvas_points_new(nodes);

    int node = 0;
    node_chain = map_item->way->node_chain;
    while(node_chain) {
      canvas_point_set_pos(points, node++, &node_chain->node->lpos);
      node_chain = node_chain->next;
    }
    
    /* create a copy of this map item and mark it as being a highlight */
    map_item_t *new_map_item = g_new0(map_item_t, 1);
    memcpy(new_map_item, map_item, sizeof(map_item_t));
    new_map_item->highlight = TRUE;
    
    map_hl_polyline_new(map, CANVAS_GROUP_WAYS_HL, new_map_item, points, 
			(map_item->way->draw_flags & OSM_DRAW_FLAGS_STROKE)?
	                 2*STROKE_WIDTH:3*WAY_WIDTH, HIGHLIGHT_COLOR);

    canvas_points_free(points);
  }
}

static void map_item_select(appdata_t *appdata, map_item_t *map_item) {
  switch(map_item->type) {
  case MAP_TYPE_NODE:
    map_node_select(appdata, map_item->node);
    break;
  case MAP_TYPE_WAY:
    map_way_select(appdata, map_item->way);
    break;
  default:
    g_assert((map_item->type == MAP_TYPE_NODE)||
	     (map_item->type == MAP_TYPE_WAY));
    break;
  }
}

static void map_item_deselect(appdata_t *appdata) {

  /* remove statusbar message */
  statusbar_set(appdata, NULL);

  /* disable/enable icons in icon bar */
  icon_bar_map_item_selected(appdata, NULL, FALSE);

  /* remove highlight */
  map_hl_remove(appdata);

  /* forget about selection */
  appdata->map->selected.type = MAP_TYPE_ILLEGAL;
}

/* called whenever a map item is to be destroyed */
static gint map_item_destroy_event(GtkWidget *widget, gpointer data) {
  map_item_t *map_item = (map_item_t*)data;

  //  printf("destroying map_item @ %p\n", map_item);

  /* remove item from nodes/ways map_item_chain */
  map_item_chain_t **chain = NULL;
  if(map_item->type == MAP_TYPE_NODE)
    chain = &map_item->node->map_item_chain;
  else if(map_item->type == MAP_TYPE_WAY)
    chain = &map_item->way->map_item_chain;

  /* there must be a chain with content, otherwise things are broken */
  g_assert(chain);
  g_assert(*chain);

  /* search current map_item, ... */
  while(*chain && (*chain)->map_item != map_item)
    chain = &(*chain)->next;

  g_assert(*chain);

  /* ... remove it from chain and free it */
  map_item_chain_t *tmp = *chain;
  *chain = (*chain)->next;

  g_free(tmp);
  g_free(map_item);

  return FALSE;
}

static canvas_item_t *map_node_new(map_t *map, node_t *node, gint radius, 
		   gint width, canvas_color_t fill, canvas_color_t border) {

  map_item_t *map_item = g_new0(map_item_t, 1);
  map_item->type = MAP_TYPE_NODE;
  map_item->node = node;
  map_item->item = canvas_circle_new(map, CANVAS_GROUP_NODES, 
		     node->lpos.x, node->lpos.y, radius, width, fill, border);
  
  /* attach map_item to nodes map_item_chain */
  map_item_chain_t **chain = &node->map_item_chain;
  while(*chain) chain = &(*chain)->next;
  *chain = g_new0(map_item_chain_t, 1);
  (*chain)->map_item = map_item;

  canvas_item_set_user_data(map_item->item, map_item);

  canvas_item_destroy_connect(map_item->item, 
          G_CALLBACK(map_item_destroy_event), map_item);

  return map_item->item;
}

/* in the rare case that a way consists of only one node, it is */
/* drawn as a circle. This e.g. happens when drawing a new way */
static canvas_item_t *map_way_single_new(map_t *map, way_t *way, gint radius, 
		   gint width, canvas_color_t fill, canvas_color_t border) {

  map_item_t *map_item = g_new0(map_item_t, 1);
  map_item->type = MAP_TYPE_WAY;
  map_item->way = way;
  map_item->item = canvas_circle_new(map, CANVAS_GROUP_WAYS, 
	  way->node_chain->node->lpos.x, way->node_chain->node->lpos.y, 
				     radius, width, fill, border);
  
  /* attach map_item to nodes map_item_chain */
  map_item_chain_t **chain = &way->map_item_chain;
  while(*chain) chain = &(*chain)->next;
  *chain = g_new0(map_item_chain_t, 1);
  (*chain)->map_item = map_item;

  canvas_item_set_user_data(map_item->item, map_item);

  canvas_item_destroy_connect(map_item->item, 
          G_CALLBACK(map_item_destroy_event), map_item);

  return map_item->item;
}

static canvas_item_t *map_way_new(map_t *map, canvas_group_t group, way_t *way, 
		  canvas_points_t *points, gint width, canvas_color_t color,
		  canvas_color_t fill_color) {
  map_item_t *map_item = g_new0(map_item_t, 1);
  map_item->type = MAP_TYPE_WAY;
  map_item->way = way;

  if(way->draw_flags & OSM_DRAW_FLAGS_POLYGON) {
    map_item->item = canvas_polygon_new(map, group, points, 
					width, color, fill_color);
  } else {
    map_item->item = canvas_polyline_new(map, group, points, width, color);
  }

  /* attach map_item to ways map_item_chain */
  map_item_chain_t **chain = &way->map_item_chain;
  while(*chain) chain = &(*chain)->next;
  *chain = g_new0(map_item_chain_t, 1);
  (*chain)->map_item = map_item;

  canvas_item_set_user_data(map_item->item, map_item);

  canvas_item_destroy_connect(map_item->item, 
	      G_CALLBACK(map_item_destroy_event), map_item);

  return map_item->item;
}

void map_show_node(map_t *map, node_t *node) {
  map_node_new(map, node, S_RADIUS, 0, NODE_WAY_COLOR, 0);
}

static void map_way_draw(map_t *map, way_t *way) {

  /* don't draw a way that's not there anymore */
  if(way->flags & OSM_FLAG_DELETED)
    return;

  /* allocate space for nodes */
  /* a way needs at least 2 points to be drawn */
  guint nodes = osm_way_number_of_nodes(way);
  if(nodes == 1) {
    /* draw a single dot where this single node is */
    map_way_single_new(map, way, S_RADIUS, 0, NODE_WAY_COLOR, 0);
  } else {
    canvas_points_t *points = canvas_points_new(nodes);
    
    int node = 0;
    node_chain_t *node_chain = way->node_chain;
    while(node_chain) {
      canvas_point_set_pos(points, node++, &node_chain->node->lpos);
      node_chain = node_chain->next;
    }
    
    /* draw way */
    if(way->draw_flags & OSM_DRAW_FLAGS_POLYGON) {
      map_way_new(map, CANVAS_GROUP_POLYGONS, way, points, 
		  2, DARK_GREY, way->fill_color);
    } else {
      map_way_new(map, CANVAS_GROUP_WAYS, way, points, 
		  WAY_WIDTH, way->color, NO_COLOR);
      
      if(way->draw_flags & OSM_DRAW_FLAGS_STROKE)
	map_way_new(map, CANVAS_GROUP_WAYS_OL, way, points, 
		    STROKE_WIDTH, DARK_GREY, NO_COLOR);
    }
    canvas_points_free(points);
  }
}

static void map_node_draw(map_t *map, node_t *node) {
  /* don't draw a node that's not there anymore */
  if(node->flags & OSM_FLAG_DELETED)
    return;
  
  if(!node->ways) 
    map_node_new(map, node, RADIUS, RADIUS/2.0, NODE_COLOR, NODE_WAY_COLOR);
  else if(osm_node_has_tag(node)) 
    map_node_new(map, node, S_RADIUS, 0, NODE_WAY_COLOR, 0);
}

static void map_item_draw(map_t *map, map_item_t *map_item) {
  switch(map_item->type) {
  case MAP_TYPE_NODE:
    map_node_draw(map, map_item->node);
    break;
  case MAP_TYPE_WAY:
    map_way_draw(map, map_item->way);
    break;
  default:
    g_assert((map_item->type == MAP_TYPE_NODE) ||
	     (map_item->type == MAP_TYPE_WAY));
  }
}

static void map_item_remove(map_t *map, map_item_t *map_item) {
  map_item_chain_t *chain = NULL;

  switch(map_item->type) {
  case MAP_TYPE_NODE:
    chain = map_item->node->map_item_chain;
    break;
  case MAP_TYPE_WAY:
    chain = map_item->way->map_item_chain;
    break;
  default:
    g_assert((map_item->type == MAP_TYPE_NODE) ||
	     (map_item->type == MAP_TYPE_WAY));
  }

  while(chain) {
    map_item_chain_t *next = chain->next;
    canvas_item_destroy(chain->map_item->item);
    chain = next;
  }
}

static void map_item_init(potlatch_conf_t *potlatch, map_item_t *map_item) {
  if(map_item->type == MAP_TYPE_WAY)
    map_way_color(potlatch, map_item->way);
}

void map_item_redraw(appdata_t *appdata, map_item_t *map_item) {
  map_item_t item = *map_item;

  /* check if the item to be redrawn is the selected one */
  gboolean is_selected = FALSE;
  if(map_item->ptr == appdata->map->selected.ptr) {
    map_item_deselect(appdata);
    is_selected = TRUE;
  }

  map_item_remove(appdata->map, &item);
  map_item_init(appdata->potlatch, &item);
  map_item_draw(appdata->map, &item);

  /* restore selection if there was one */
  if(is_selected) 
    map_item_select(appdata, &item);
}

static void map_draw(map_t *map, osm_t *osm) {
  g_assert(map->canvas);

  printf("drawing ways ...\n");
  way_t *way = osm->way;
  while(way) {
    map_way_draw(map, way);
    way = way->next;
  }

  printf("drawing single nodes ...\n");
  node_t *node = osm->node;
  while(node) {
    map_node_draw(map, node);
    node = node->next;
  }
}

static gint map_destroy_event(GtkWidget *widget, gpointer data) {
  appdata_t *appdata = (appdata_t*)data;
  map_t *map = appdata->map;

  printf("destroying map\n");

  g_free(map);
  appdata->map = NULL;

  return FALSE;
}

/* get the item at position x, y */
static map_item_t *map_item_at(map_t *map, gint x, gint y) {	  
  printf("map check at %d/%d\n", x, y);

  canvas_window2world(map->canvas, x, y, &x, &y);

  printf("world check at %d/%d\n", x, y);
  
  canvas_item_t *item = canvas_get_item_at(map->canvas, x, y);
	  
  if(!item) {
    printf("  there's no item\n");
    return NULL;
  }

  printf("  there's an item (%p)\n", item);

  map_item_t *map_item = canvas_item_get_user_data(item);

  if(!map_item) {
    printf("  item has no user data!\n");
    return NULL;
  }
    
  if(map_item->highlight)
    printf("  item is highlight\n");    

  switch(map_item->type) {
  case MAP_TYPE_NODE:
    printf("  item is node #%ld\n", map_item->node->id);
    break;
  case MAP_TYPE_WAY:
    printf("  item is way #%ld\n", map_item->way->id);
    break;
  default:
    printf("  unknown item\n");
    break;
  }

  return map_item;
}    

/* get the real item (no highlight) at x, y */
static map_item_t *map_real_item_at(map_t *map, gint x, gint y) { 
  map_item_t *map_item = map_item_at(map, x, y);

  /* no item or already a real one */
  if(!map_item || !map_item->highlight) return map_item;

  /* get the item (parent) this item is the highlight of */
  map_item_t *parent = NULL;
  switch(map_item->type) {

  case MAP_TYPE_NODE:
    if(map_item->node->map_item_chain)
      parent = map_item->node->map_item_chain->map_item;

    if(parent)
      printf("  using parent item node #%ld\n", parent->node->id);      
    break;

  case MAP_TYPE_WAY:
    if(map_item->way->map_item_chain)
      parent = map_item->way->map_item_chain->map_item;

    if(parent)
      printf("  using parent item way #%ld\n", parent->way->id);      
    break;

  default:
    g_assert((map_item->type == MAP_TYPE_NODE) ||
	     (map_item->type == MAP_TYPE_WAY)); 
    break;
  }
  
  if(parent) 
    map_item = parent;
  else 
    printf("  no parent, working on highlight itself\n");

  return map_item;
}

void map_set_zoom(map_t *map, double zoom) {
  map->zoom = zoom;
  canvas_set_zoom(map->canvas, map->zoom);
}

static gboolean map_scroll_event(GtkWidget *widget, GdkEventScroll *event,
				 gpointer data) {
  appdata_t *appdata = (appdata_t*)data;

  if(event->type == GDK_SCROLL) {
    if(event->direction) 
      map_set_zoom(appdata->map, appdata->map->zoom / ZOOM_FACTOR_WHEEL);
    else
      map_set_zoom(appdata->map, appdata->map->zoom * ZOOM_FACTOR_WHEEL);
  }

  return TRUE;
}

static gboolean distance_above(map_t *map, gint x, gint y, gint limit) {
  gint sx, sy;

#ifdef USE_GNOMECANVAS
  gnome_canvas_get_scroll_offsets(GNOME_CANVAS(map->canvas), &sx, &sy);

  /* add offsets generated by mouse within map and map scrolling */
  sx = (x-map->pen_down.at.x) + (map->pen_down.so.x-sx);
  sy = (y-map->pen_down.at.y) + (map->pen_down.so.y-sy);
#else
  /* add offsets generated by mouse within map and map scrolling */
  sx = (x-map->pen_down.at.x);
  sy = (y-map->pen_down.at.y);
#endif

  return(sx*sx + sy*sy > limit*limit);
}

/* scroll with respect to two screen positions */
static void map_do_scroll(map_t *map, gint x, gint y) {
  gint sx, sy;

  canvas_get_scroll_offsets(map->canvas, &sx, &sy);
  sx -= x-map->pen_down.at.x;
  sy -= y-map->pen_down.at.y;
  canvas_scroll_to(map->canvas, sx, sy);
}

/* scroll a certain step */
static void map_do_scroll_step(map_t *map, gint x, gint y) {
  gint sx, sy;
  canvas_get_scroll_offsets(map->canvas, &sx, &sy);
  canvas_scroll_to(map->canvas, sx+x, sy+y);
}


static gboolean item_is_selected_node(map_t *map, map_item_t *map_item) {
  printf("check if item is a selected node\n");

  if(map->selected.type == MAP_TYPE_ILLEGAL) {
    printf("  nothing is selected\n");
    return FALSE;
  }

  /* clicked the highlight directly */
  if(map_item->type != MAP_TYPE_NODE) {
    printf("  didn't click node\n");
    return FALSE;
  }

  if(map->selected.type == MAP_TYPE_NODE) {
    printf("  selected item is a node\n");

    if(map_item->node == map->selected.node) {
      printf("  requested item is a selected node\n");
      return TRUE;
    }
    printf("  but it's not the requested one\n");
    return FALSE;

  } else if(map->selected.type == MAP_TYPE_WAY) {
    printf("  selected item is a way\n");

    node_chain_t *node_chain = map->selected.way->node_chain;
    while(node_chain) {
      if(node_chain->node == map_item->node) {
	printf("  requested item is part of selected way\n");
	return TRUE;
      } 
      node_chain = node_chain->next;
    }
    printf("  but it doesn't include the requested node\n");
    return FALSE;
    
  } else {
    printf("  selected item is unknown\n");
    return FALSE;
  }

  g_assert(0);
  return TRUE;
}

/* return true if the item given is the currenly selected way */
/* also return false if nothing is selected or the selection is no way */
static gboolean item_is_selected_way(map_t *map, map_item_t *map_item) {
  printf("check if item is the selected way\n");

  if(!map_item) {
    printf("  no item requested\n");
    return FALSE;
  }

  if(map->selected.type == MAP_TYPE_ILLEGAL) {
    printf("  nothing is selected\n");
    return FALSE;
  }

  /* clicked the highlight directly */
  if(map_item->type != MAP_TYPE_WAY) {
    printf("  didn't click way\n");
    return FALSE;
  }

  if(map->selected.type == MAP_TYPE_WAY) {
    printf("  selected item is a way\n");

    if(map_item->way == map->selected.way) {
      printf("  requested item is a selected way\n");
      return TRUE;
    }
    printf("  but it's not the requested one\n");
    return FALSE;
  }

  printf("  selected item is not a way\n");
  return FALSE;
}


void map_highlight_refresh(appdata_t *appdata) {
  map_t *map = appdata->map;
  map_item_t old = map->selected;

  printf("type to refresh is %d\n", old.type);
  if(old.type == MAP_TYPE_ILLEGAL) 
    return;

  map_item_deselect(appdata);
  map_item_select(appdata, &old);
}

static void map_way_delete(appdata_t *appdata, way_t *way) {
  printf("deleting way #%ld from map and osm\n", way->id);

  /* remove it visually from the screen */
  map_item_chain_destroy(&way->map_item_chain);

  /* also remove the visible representation of all nodes (nodes with tags) */
  node_chain_t *chain = way->node_chain;
  while(chain) {
    if(chain->node->map_item_chain) 
      map_item_chain_destroy(&chain->node->map_item_chain);
      
    chain = chain->next;
  }

  /* and mark it "deleted" in the database */
  osm_way_remove_from_relation(appdata->osm, way);
  osm_way_delete(appdata->osm, way, FALSE);
}

static void map_node_move(appdata_t *appdata, map_item_t *map_item, 
			  gint ex, gint ey) {

  map_t *map = appdata->map;
  osm_t *osm = appdata->osm;

  g_assert(map_item->type == MAP_TYPE_NODE);
  node_t *node = map_item->node;

  printf("released dragged node #%ld\n", node->id);
  printf("  was at %d %d (%f %f)\n", 
	 node->lpos.x, node->lpos.y,
	 node->pos.lat, node->pos.lon);


  /* check if it was dropped onto another node */
  node_t *touchnode = map_hl_touchnode_get_node(map);
  if(touchnode) {
    map_hl_touchnode_clear(map);

    printf("  dropped onto node #%ld\n", touchnode->id);

    if(yes_no_f(GTK_WIDGET(appdata->window), _("Join nodes?"),
		_("Do you want to join the dragged node with the one "
		  "you dropped it on?"))) {

      /* the touchnode vanishes and is replaced by the node the */
      /* user dropped onto it */

      /* TODO: move this into a function in osm.c */
      way_t *way = appdata->osm->way;
      while(way) {
	node_chain_t *chain = way->node_chain;
	while(chain) {
	  if(chain->node == touchnode) {
	    printf("  found node in way #%ld\n", way->id);
	    
	    /* replace by node */
	    chain->node = node;
	    
	    /* and adjust way references of both nodes */
	    node->ways++;
	    touchnode->ways--;
	    
	    way->flags |= OSM_FLAG_DIRTY;
	  }
	  chain = chain->next;
	}
	way = way->next;
      }
      
      /* replace node in relations */
      relation_t *relation = appdata->osm->relation;
      while(relation) {
	member_t *member = relation->member;
	while(member) {
	  if(member->type == NODE && member->node == touchnode) {
	    printf("  found node in relation #%ld\n", relation->id);
	    
	    /* replace by node */
	    member->node = node;
	    
	    relation->flags |= OSM_FLAG_DIRTY;
	  }
	  member = member->next;
	}
	relation = relation->next;
      }

      /* transfer tags from touchnode to node */
      tag_t **dst = &node->tag;
      while(*dst) dst = &((*dst)->next);  /* find end of target tag list */
      tag_t **src = &touchnode->tag;
      gboolean conflict = FALSE;
      while(*src) {
	/* check if same key but with different value is present, */
	/* ignoring the created_by tags */
	if(!osm_is_creator_tag(*src) && 
	   osm_tag_key_other_value_present(node->tag, *src))
	  conflict = TRUE;

	/* don't copy "created_by" and "source" tag or tags that already */
	/* exist in identical form */
	if(osm_is_creator_tag(*src) || 
	   osm_tag_key_and_value_present(node->tag, *src)) {
	  tag_t *next = (*src)->next;
	  osm_tag_free(*src);
	  *src = next;
	} else {
	  *dst = *src;
	  *src = (*src)->next;
	  dst = &((*dst)->next);
	  *dst = NULL;
	}
      }

      /* touchnode must not have any references to ways anymore */
      g_assert(!touchnode->ways);

      /* delete touchnode */
      /* remove it visually from the screen */
      map_item_chain_destroy(&touchnode->map_item_chain);

      /* and remove it from the data structures */
      osm_node_remove_from_relation(appdata->osm, touchnode);
      osm_node_delete(appdata->osm, touchnode, FALSE, TRUE);
      
      /* and open dialog to resolve tag collisions if necessary */
      if(conflict) {
	messagef(GTK_WIDGET(appdata->window), "Node tag conflict", 
		 "The resulting node contains some conflicting tags. "
		 "Please solve these.");
      }
    
      /* check whether this will also join two ways */
      printf("  checking if node is end of way\n");
      guint ways2join_cnt = 0;
      way_t *ways2join[2] = { NULL, NULL };
      way = appdata->osm->way;
      while(way) {
	if(osm_way_ends_with_node(way, node)) {
	  if(ways2join_cnt < 2) 
	    ways2join[ways2join_cnt] = way;

	  printf("  way #%ld ends with this node\n", way->id);
	  ways2join_cnt++;
	}
	way = way->next;
      }

      if(ways2join_cnt > 2) {
	messagef(GTK_WIDGET(appdata->window), _("Too many ways to join"), 
		 _("More than two ways now end in this node. Joining more "
		   "than two ways is not yet implemented, sorry"));
	
      } else if(ways2join_cnt == 2) {
	if(yes_no_f(GTK_WIDGET(appdata->window), 
		    _("Join ways?"),
		    _("Do you want to join the dragged way with the one "
		      "you dropped it on?"))) {

	  printf("  about to join ways #%ld and #%ld\n",
		 ways2join[0]->id, ways2join[1]->id);

	  /* way[1] gets destroyed and attached to way[0] */
	  /* so check if way[1] is selected and exchainge ways then */
	  /* so that way may stay selected */
	  if((map->selected.type == MAP_TYPE_WAY) && 
	     (map->selected.way == ways2join[1])) {
	    printf("  swapping ways to keep selected one alive\n");
	    way_t *tmp = ways2join[1];
	    ways2join[1] = ways2join[0];
	    ways2join[0] = tmp;
	  }

	  /* take all nodes from way[1] and append them to way[0] */
	  /* check if we have to append or prepend to way[0] */
	  gboolean revert = FALSE;
	  if(ways2join[0]->node_chain->node == node) {
	    /* make "prepend" to be "append" by reverting way[0] */
	    printf("  target prepend -> revert\n");	    
	    revert = TRUE;
	    osm_way_revert(ways2join[0]);
	  } 
	  
	  /* verify the common node is last in the target way */
	  node_chain_t *chain = ways2join[0]->node_chain;
	  while(chain->next) chain = chain->next;
	  g_assert(chain->node == node);
	  g_assert(!chain->next);
	  
	  /* common node must be first in the chain to attach */
	  if(ways2join[1]->node_chain->node != node) {
	    printf("  source revert\n");	    
	    osm_way_revert(ways2join[1]);
	  }
	  
	  /* verify the common node is first in the source way */
	  g_assert(ways2join[1]->node_chain->node == node);

	  /* finally append source chain to target */
	  g_assert(!chain->next);
	  chain->next = ways2join[1]->node_chain->next;

	  ways2join[1]->node_chain->next = NULL;

	  /* ---------- transfer tags from way[1] to way[0] ----------- */
	  tag_t **dst = &ways2join[0]->tag;
	  while(*dst) dst = &((*dst)->next);  /* find end of target tag list */
	  tag_t **src = &ways2join[1]->tag;
	  gboolean conflict = FALSE;
	  while(*src) {
	    /* check if same key but with different value is present, */
	    /* ignoring the created_by tags */
	    if(!osm_is_creator_tag(*src) && 
	       osm_tag_key_other_value_present(ways2join[0]->tag, *src))
	      conflict = TRUE;

	    /* don't copy "created_by" and "source" tag or tags that already */
	    /* exist in identical form */
	    if(osm_is_creator_tag(*src) || 
	       osm_tag_key_and_value_present(ways2join[0]->tag, *src)) {
	      tag_t *next = (*src)->next;
	      osm_tag_free(*src);
	      *src = next;
	    } else {
	      *dst = *src;
	      *src = (*src)->next;
	      dst = &((*dst)->next);
	      *dst = NULL;
	    }
	  }

	  /* ---- transfer relation membership from way[1] to way[0] ----- */
	  relation_chain_t *rchain = 
	    osm_way_to_relation(appdata->osm, ways2join[1]);

	  while(rchain) {
	    relation_chain_t *next = rchain->next;
	    printf("way[1] is part of relation #%ld\n", rchain->relation->id);

	    /* make way[0] member of the same relation */

	    /* walk member chain. save role of way[1] if its being found. */
	    /* end search either at end of chain or if way[0] was foind */
	    /* as it's already a member of that relation */
	    member_t **member = &rchain->relation->member;
	    char *role = NULL;
	    while(*member && 
		  !(((*member)->type == WAY) && 
		    ((*member)->way == ways2join[0]))) {

	      /* save role of way[1] */
	      if(((*member)->type == WAY) && ((*member)->way == ways2join[0]))
		role = (*member)->role;

	      member = &(*member)->next;
	    }

	    if(*member) 
	      printf("  both ways were members of this relation\n");
	    else {
	      printf("  adding way[0] to relation\n");
	      *member = g_new0(member_t, 1);
	      (*member)->type = WAY;
	      (*member)->way = ways2join[0];
	      if(role) (*member)->role = g_strdup(role);
	      member = &(*member)->next;

	      rchain->relation->flags |= OSM_FLAG_DIRTY;
	    }

	    g_free(rchain);
	    rchain = next;
	  }

	  
	  /* and open dialog to resolve tag collisions if necessary */
	  if(conflict) {
	    messagef(GTK_WIDGET(appdata->window), "Way tag conflict", 
		     "The resulting way contains some conflicting tags. "
		     "Please solve these.");
	  }
	  
	  ways2join[0]->flags |= OSM_FLAG_DIRTY;
	  map_way_delete(appdata, ways2join[1]);
	}
      }
    }
  }

  /* finally update dragged nodes position */

  /* convert mouse position to canvas (world) position */
  gint x, y;
  canvas_window2world(map->canvas, ex, ey, &x, &y);
  if(!osm_position_within_bounds(appdata->osm, x, y)) {
    map_outside_error(appdata);
    return;
  }
  
  node->lpos.x = x;
  node->lpos.y = y;

  /* convert mouse position to canvas (world) position */
  canvas_window2world(map->canvas, ex, ey, &node->lpos.x, &node->lpos.y);
	    
  /* convert screen position back to ll */
  lpos2pos(osm->bounds, &node->lpos, &node->pos);

  printf("  now at %d %d (%f %f)\n", 
	 node->lpos.x, node->lpos.y, node->pos.lat, node->pos.lon);
  
  /* now update the visual representation of the node */
  canvas_item_t *item = map_item->item;
  
  if(!node->ways) 
    canvas_item_set_pos(item, &node->lpos, RADIUS);
  else {
    /* only move item if it's not the highlight. highlight will be */
    /* redrawed entirely lateron */
    if(!map_item->highlight)
      canvas_item_set_pos(item, &node->lpos, S_RADIUS);
  }

  /* visually update ways, node is part of */
  way_t *way = osm->way;
  while(way) {
    if(osm_node_in_way(way, node)) {
      printf("  node is part of way #%ld, redraw!\n", way->id);
      
      /* remove prior version of this way */
      map_item_chain_destroy(&way->map_item_chain);

      /* draw current way */
      map_way_color(appdata->potlatch, way);
      map_way_draw(map, way);
    }
    
    way = way->next;
  }
  
  /* and mark the node as dirty */
  node->flags |= OSM_FLAG_DIRTY;

  /* update highlight */
  map_highlight_refresh(appdata);
}
			  
static void map_handle_click(appdata_t *appdata, map_t *map) {

  /* problem: on_item may be the highlight itself! So store it! */
  map_item_t map_item;
  if(map->pen_down.on_item) map_item = *map->pen_down.on_item;
  else                      map_item.type = MAP_TYPE_ILLEGAL;

  /* if we aready have something selected, then de-select it */
  map_item_deselect(appdata);

  /* select the clicked item (if there was one) */
  if(map_item.type != MAP_TYPE_ILLEGAL) {
    switch(map_item.type) {
    case MAP_TYPE_NODE:
      map_node_select(appdata, map_item.node);
      break;

    case MAP_TYPE_WAY:
      map_way_select(appdata, map_item.way);
      break;

    default:
      g_assert((map_item.type == MAP_TYPE_NODE) ||
	       (map_item.type == MAP_TYPE_WAY));
      break;
    }
  }
}

static void map_touchnode_update(appdata_t *appdata, gint x, gint y) {
  map_t *map = appdata->map;

  map_hl_touchnode_clear(map);

  node_t *cur_node = NULL;

  /* the "current node" which is the one we are working on and which */
  /* should not be highlighted depends on the action */
  switch(map->action.type) {

    /* in idle mode the dragged node is not highlighted */
  case MAP_ACTION_IDLE:
    g_assert(map->pen_down.on_item);
    g_assert(map->pen_down.on_item->type == MAP_TYPE_NODE);
    cur_node = map->pen_down.on_item->node;
    break;

  default:
    break;
  }


  /* check if we are close to one of the other nodes */
  canvas_window2world(appdata->map->canvas, x, y, &x, &y);
  node_t *node = appdata->osm->node;
  while(!map->touchnode && node) {
    /* don't highlight the dragged node itself and don't highlight */
    /* deleted ones */
    if((node != cur_node) && (!(node->flags & OSM_FLAG_DELETED))) {
      gint nx = x - node->lpos.x;
      gint ny = y - node->lpos.y;

      if(nx*nx + ny*ny < RADIUS*RADIUS)
	map_hl_touchnode_draw(map, node);
    }
    node = node->next;
  }

  /* during way creation also nodes of the new way */
  /* need to be searched */
  if(!map->touchnode && map->action.way) {
    node_chain_t *chain = map->action.way->node_chain;
    while(!map->touchnode && chain && chain->next) {
      gint nx = x - chain->node->lpos.x;
      gint ny = y - chain->node->lpos.y;
	  
      if(nx*nx + ny*ny < RADIUS*RADIUS)
	map_hl_touchnode_draw(map, chain->node);

      chain = chain->next;
    }
  }
}

static void map_button_press(map_t *map, gint x, gint y) {

  printf("left button pressed\n");
  map->pen_down.is = TRUE;
      
  /* save press position */
  map->pen_down.at.x = x;
  map->pen_down.at.y = y;
  map->pen_down.drag = FALSE;     // don't assume drag yet

#ifdef USE_GNOMECANVAS  
  /* save initial scroll offset */
  gnome_canvas_get_scroll_offsets(GNOME_CANVAS(map->canvas), 
		  &map->pen_down.so.x, &map->pen_down.so.y);
#endif  

  /* determine wether this press was on an item */
  map->pen_down.on_item = map_real_item_at(map, x, y);
  
  /* check if the clicked item is a highlighted node as the user */
  /* might want to drag that */
  map->pen_down.on_selected_node = FALSE;
  if(map->pen_down.on_item)
    map->pen_down.on_selected_node = 
      item_is_selected_node(map, map->pen_down.on_item);

  /* button press */
  switch(map->action.type) {

  case MAP_ACTION_WAY_NODE_ADD:
  case MAP_ACTION_WAY_CUT:
    if(item_is_selected_way(map, map->pen_down.on_item)) {
      gint nx, ny;
      canvas_window2world(map->canvas, x, y, &nx, &ny);
      if(canvas_item_get_segment(map->pen_down.on_item->item, nx, ny) >= 0)
	map_hl_cursor_draw(map, x, y);
    }
    break;

  case MAP_ACTION_NODE_ADD:
    map_hl_cursor_draw(map, x, y);
    break;

  case MAP_ACTION_WAY_ADD:
    map_hl_cursor_draw(map, x, y);
    map_touchnode_update(map->appdata, x, y);
    break;

  default:
    break;
  }
}

/* move the background image (wms data) during wms adjustment */
static void map_bg_adjust(map_t *map, gint x, gint y) {    
  g_assert(map->appdata);
  g_assert(map->appdata->osm);
  g_assert(map->appdata->osm->bounds);

  x += map->appdata->osm->bounds->min.x + map->bg_offset.x -
    map->pen_down.at.x;
  y += map->appdata->osm->bounds->min.y + map->bg_offset.y -
    map->pen_down.at.y;

  canvas_image_move(map->bg_item, x, y, map->bg_scale.x, map->bg_scale.y);
}

static void map_button_release(map_t *map, gint x, gint y) {
  map->pen_down.is = FALSE;

  /* before button release is handled */
  switch(map->action.type) {
  case MAP_ACTION_BG_ADJUST:
    map_bg_adjust(map, x, y);
    map->bg_offset.x += x - map->pen_down.at.x;
    map->bg_offset.y += y - map->pen_down.at.y;
    break;

  case MAP_ACTION_IDLE:
    /* check if distance to press is above drag limit */
    if(!map->pen_down.drag)
      map->pen_down.drag = distance_above(map, x, y, MAP_DRAG_LIMIT);
    
    if(!map->pen_down.drag) {
      printf("left button released after click\n");
      
      map_item_t old_sel = map->selected;
      map_handle_click(map->appdata, map);
      
      if((old_sel.type != MAP_TYPE_ILLEGAL) && 
	 (old_sel.type == map->selected.type) &&
	 (old_sel.ptr == map->selected.ptr)) {
	printf("re-selected same item of type %d, "
	       "pushing it to the bottom\n", old_sel.type);
	
	if(!map->selected.item) {
	  printf("  item has no visible representation to push\n");
	} else {
	  canvas_item_to_bottom(map->selected.item);
	  
	  /* update clicked item, to correctly handle the click */
	  map->pen_down.on_item = 
	    map_real_item_at(map, map->pen_down.at.x, map->pen_down.at.y);
	  
	  map_handle_click(map->appdata, map);
	}
      }
    } else {
      printf("left button released after drag\n");
      
      /* just scroll if we didn't drag an selected item */
      if(!map->pen_down.on_selected_node)
	map_do_scroll(map, x, y);
      else {
	printf("released after dragging node\n");
	map_hl_cursor_clear(map);
	
	/* now actually move the node */
	map_node_move(map->appdata, map->pen_down.on_item, x, y);
      }
    }
    break;

  case MAP_ACTION_NODE_ADD:
    printf("released after NODE ADD\n");
    map_hl_cursor_clear(map);

    /* convert mouse position to canvas (world) position */
    canvas_window2world(map->canvas, x, y, &x, &y);

    node_t *node = NULL;
    if(!osm_position_within_bounds(map->appdata->osm, x, y)) 
      map_outside_error(map->appdata);
    else {
      node = osm_node_new(map->appdata->osm, x, y);
      osm_node_attach(map->appdata->osm, node);
      map_node_draw(map, node);
    }
    map_action_set(map->appdata, MAP_ACTION_IDLE);
      
    map_item_deselect(map->appdata);

    if(node) {
      map_node_select(map->appdata, node);

      /* let the user specify some tags for the new node */
      info_dialog(map->appdata);
    }
    break;

  case MAP_ACTION_WAY_ADD:
    printf("released after WAY ADD\n");
    map_hl_cursor_clear(map);

    /* convert mouse position to canvas (world) position */
    canvas_window2world(map->canvas, x, y, &x, &y);

    /* check if this was a double click. This is the case if */
    /* the last node placed is less than 5 pixels from the current */
    /* position */
    node_t *lnode = osm_way_get_last_node(map->action.way);
    if(lnode && (map->zoom * sqrt((lnode->lpos.x-x)*(lnode->lpos.x-x)+ 
				  (lnode->lpos.y-y)*(lnode->lpos.y-y))) < 5) {
      printf("detected double click -> simulate ok click\n");
      map_action_ok(map->appdata);
    } else {
      
      /* use the existing node if one was touched */
      node = map_hl_touchnode_get_node(map);
      if(node) {
	printf("  re-using node #%ld\n", node->id);
	map_hl_touchnode_clear(map);
      } else {
	if(!osm_position_within_bounds(map->appdata->osm, x, y)) 
	  map_outside_error(map->appdata);
	else 
	  node = osm_node_new(map->appdata->osm, x, y);
      }
      
      if(node) {
	g_assert(map->action.way);
	osm_way_append_node(map->action.way, node);
	
	switch(osm_node_chain_length(map->action.way->node_chain)) {
	case 1:
	  /* replace "place first node..." message */
	  statusbar_set(map->appdata, _("Place next node of way"));
	  break;
	  
	case 2:
	  /* two nodes are enough for a valid way */
	  icon_bar_map_cancel_ok(map->appdata, TRUE, TRUE);
	  break;
	  
	default:
	  break;
	}
	
	/* remove prior version of this way */
	map_item_chain_destroy(&map->action.way->map_item_chain);
	
	/* draw current way */
	map_way_color(map->appdata->potlatch, map->action.way);
	map_way_draw(map, map->action.way);
      }
    }
    break;

  case MAP_ACTION_WAY_NODE_ADD:
    printf("released after WAY NODE ADD\n");
    map_hl_cursor_clear(map);

    /* check if we are still hovering above the selected way */
    map_item_t *item = map_item_at(map, x, y);
    if(item && item_is_selected_way(map, item)) {
      /* convert mouse position to canvas (world) position */
      canvas_window2world(map->canvas, x, y, &x, &y);
      gint insert_after = canvas_item_get_segment(item->item, x, y)+1;
      if(insert_after > 0) {
	/* create new node */
	node_t* node = osm_node_new(map->appdata->osm, x, y);
	osm_node_attach(map->appdata->osm, node);

	/* insert it into ways chain of nodes */
	way_t *way = item->way;

	/* search correct position */
	node_chain_t **chain = &way->node_chain;
	while(insert_after--) {
	  g_assert(*chain);
	  chain = &(*chain)->next;
	}

	/* and actually do the insertion */
	node_chain_t *new_chain_item = g_new0(node_chain_t, 1);
	new_chain_item->node = node;
	new_chain_item->next = *chain;
	*chain = new_chain_item;

	/* clear selection */
	map_item_deselect(map->appdata);

	/* remove prior version of this way */
	map_item_chain_destroy(&way->map_item_chain);

	/* draw the updated way */
	map_way_draw(map, way);

	/* remember that this node is contained in one way */
	node->ways=1;

	/* and that the way needs to be uploaded */
	way->flags |= OSM_FLAG_DIRTY;

	/* put gui into idle state */
	map_action_set(map->appdata, MAP_ACTION_IDLE);

	/* and redo it */
       	map_way_select(map->appdata, way);
      }
    }
    break;

  case MAP_ACTION_WAY_CUT:
    printf("released after WAY CUT\n");
    map_hl_cursor_clear(map);

    /* check if we are still hovering above the selected way */
    item = map_item_at(map, x, y);
    if(item && item_is_selected_way(map, item)) {
      /* convert mouse position to canvas (world) position */
      canvas_window2world(map->canvas, x, y, &x, &y);
      gint insert_after = canvas_item_get_segment(item->item, x, y);
      if(insert_after >= 0) {
	/* create a duplicate of the currently selected way */
	way_t *way = item->way;
	way_t *new = osm_way_new();

	/* if this is a closed way, reorder (rotate) it, so the */
	/* place to cut is adjecent to the begin/end of the way. */
	/* this prevents a cut polygon to be split into two ways */
	g_assert(way->node_chain);
	if(way->node_chain->node == osm_way_get_last_node(way)) {
	  printf("CLOSED WAY -> rotate by %d\n", insert_after);
	  osm_way_rotate(way, insert_after);
	  insert_after = 0;
	}

	/* ------------  copy all tags ------------- */
	tag_t **dst_tag = &new->tag;
	tag_t *src_tag = way->tag;
	while(src_tag) {
	  *dst_tag = g_new0(tag_t, 1);
	  (*dst_tag)->key = g_strdup(src_tag->key);
	  (*dst_tag)->value = g_strdup(src_tag->value);
	  dst_tag = &(*dst_tag)->next;
	  src_tag = src_tag->next;
	}

	/* ---- transfer relation membership from way to new ----- */
	relation_chain_t *rchain = osm_way_to_relation(map->appdata->osm, way);

	while(rchain) {
	  relation_chain_t *next = rchain->next;
	  printf("way is part of relation #%ld\n", rchain->relation->id);

	  /* make new member of the same relation */

	  /* walk member chain. save role of way if its being found. */
	  member_t **member = &rchain->relation->member;
	  char *role = NULL;
	  while(*member) { 
	    /* save role of way */
	    if(((*member)->type == WAY) && ((*member)->way == way))
	      role = (*member)->role;
	    member = &(*member)->next;
	  }

	  printf("  adding new to relation\n");
	  *member = g_new0(member_t, 1);
	  (*member)->type = WAY;
	  (*member)->way = new;
	  if(role) (*member)->role = g_strdup(role);
	  member = &(*member)->next;
	  
	  rchain->relation->flags |= OSM_FLAG_DIRTY;

	  g_free(rchain);
	  rchain = next;
	}

	/* move parts of node_chain to the new way */
	node_chain_t *chain = way->node_chain;
	while(insert_after--) {
	  g_assert(chain);
	  chain = chain->next;
	}
	 
	/* attach remaining nodes to new way */
	new->node_chain = chain->next;

	/* terminate remainig chain on old way */
	chain->next = NULL;

	/* now move the way itself into the main data structure */
	osm_way_attach(map->appdata->osm, new);
	
	/* clear selection */
	map_item_deselect(map->appdata);

	/* remove prior version of this way */
	printf("remove visible version of way #%ld\n", way->id);
	map_item_chain_destroy(&way->map_item_chain);

	/* swap chains of the old way is to be destroyed due to a lack */
	/* of nodes */
	if(osm_way_number_of_nodes(way) < 2) {
	  printf("swapping ways to avoid destruction of original way\n");
	  node_chain_t *tmp = way->node_chain;
	  way->node_chain = new->node_chain;
	  new->node_chain = tmp;
	}

	/* the way may still only consist of a single node. */
	/* remove it then */
	if(osm_way_number_of_nodes(way) < 2) {
	  printf("original way has less than 2 nodes left, deleting it\n");
	  map_way_delete(map->appdata, way);
	  item = NULL;
	} else {
	  printf("original way still has %d nodes\n",
		 osm_way_number_of_nodes(way));

	  /* draw the updated old way */
	  map_way_color(map->appdata->potlatch, way);
	  map_way_draw(map, way);

	  /* remember that the way needs to be uploaded */
	  way->flags |= OSM_FLAG_DIRTY;
	}

	if(osm_way_number_of_nodes(new) < 2) {
	  printf("new way has less than 2 nodes, deleting it\n");
	  map_way_delete(map->appdata, new);
	  new = NULL;
	} else {

	  /* colorize the new way before drawing */
	  map_way_color(map->appdata->potlatch, new);
	  map_way_draw(map, new);
	}

	/* put gui into idle state */
	map_action_set(map->appdata, MAP_ACTION_IDLE);      

	/* and redo selection if way still exists */
	if(item) 
	  map_way_select(map->appdata, way);
	else if(new)
	  map_way_select(map->appdata, new);
      }
    }

    break;


  default:
    break;
  }
}

static gboolean map_button_event(GtkWidget *widget, GdkEventButton *event,
				       gpointer data) {
  appdata_t *appdata = (appdata_t*)data;
  map_t *map = appdata->map;

  if(event->button == 1) {
    gint x = event->x, y = event->y;

    if(event->type == GDK_BUTTON_PRESS) 
      map_button_press(map, x, y);

    if(event->type == GDK_BUTTON_RELEASE) 
      map_button_release(map, x, y);
  }

  return FALSE;  /* forward to further processing */
}

static gboolean map_motion_notify_event(GtkWidget *widget, 
                             GdkEventMotion *event, gpointer data) {
  appdata_t *appdata = (appdata_t*)data;
  map_t *map = appdata->map;
  gint x, y;
  GdkModifierType state;

#ifdef USE_HILDON
  /* reduce update frequency on hildon to keep screen update fluid */
  static guint32 last_time = 0;

  if(event->time - last_time < 100) return FALSE;
  last_time = event->time;
#endif

  if(!map->pen_down.is) 
    return FALSE;

#ifdef USE_GNOMECANVAS
  /* handle hints, hints are handled by goocanvas directly */
  if(event->is_hint)
    gdk_window_get_pointer(event->window, &x, &y, &state);
  else 
#endif
  {
    x = event->x;
    y = event->y;
    state = event->state;
  }

  /* check if distance to press is above drag limit */
  if(!map->pen_down.drag)
    map->pen_down.drag = distance_above(map, x, y, MAP_DRAG_LIMIT);
  
  /* drag */
  switch(map->action.type) {
  case MAP_ACTION_BG_ADJUST:
    map_bg_adjust(map, x, y);
    break;

  case MAP_ACTION_IDLE:
    if(map->pen_down.drag) {
      /* just scroll if we didn't drag an selected item */
      if(!map->pen_down.on_selected_node)
	map_do_scroll(map, x, y);
      else {
	map_hl_cursor_draw(map, x, y);
	map_touchnode_update(appdata, x, y);
      }
    }
    break;
    
  case MAP_ACTION_NODE_ADD:
    map_hl_cursor_draw(map, x, y);
    break;
    
  case MAP_ACTION_WAY_ADD:
    map_hl_cursor_draw(map, x, y);
    map_touchnode_update(appdata, x, y);
    break;

  case MAP_ACTION_WAY_NODE_ADD:
  case MAP_ACTION_WAY_CUT:
    map_hl_cursor_clear(map);
    map_item_t *item = map_item_at(map, x, y);
    if(item && item_is_selected_way(map, item)) {
      gint nx, ny;
      canvas_window2world(map->canvas, x, y, &nx, &ny);
      if(canvas_item_get_segment(item->item, nx, ny) >= 0)
	map_hl_cursor_draw(map, x, y);
    }
    break;

  default:
    break;
  }


  return FALSE;  /* forward to further processing */
}

gboolean map_key_press_event(appdata_t *appdata, GdkEventKey *event) {
  /* map needs to be there to handle buttons */
  if(!appdata->map->canvas) 
    return FALSE;

  if(event->type == GDK_KEY_PRESS) {
    switch(event->keyval) {

    case GDK_Left:
      map_do_scroll_step(appdata->map, -50, 0);
      break;

    case GDK_Right:
      map_do_scroll_step(appdata->map, +50, 0);
      break;

    case GDK_Up:
      map_do_scroll_step(appdata->map, 0, -50);
      break;

    case GDK_Down:
      map_do_scroll_step(appdata->map, 0, +50);
      break;

    case GDK_Return:   // same as HILDON_HARDKEY_SELECT
      /* if the ok button is enabled, call its function */
      if(GTK_WIDGET_FLAGS(appdata->iconbar->ok) & GTK_SENSITIVE) 
	map_action_ok(appdata);
      /* otherwise if info is enabled call that */
      else if(GTK_WIDGET_FLAGS(appdata->iconbar->info) & GTK_SENSITIVE)
	info_dialog(appdata);
      break;

    case GDK_Escape:   // same as HILDON_HARDKEY_ESC
      /* if the cancel button is enabled, call its function */
      if(GTK_WIDGET_FLAGS(appdata->iconbar->cancel) & GTK_SENSITIVE) 
	map_action_cancel(appdata);
      break;

#ifdef USE_HILDON
    case HILDON_HARDKEY_INCREASE:
#else
    case '+':
#endif
      map_set_zoom(appdata->map, 
		   appdata->map->zoom*ZOOM_FACTOR_BUTTON);
      printf("zoom is now %f\n", appdata->map->zoom);
      return TRUE;
      break;

#ifdef USE_HILDON
    case HILDON_HARDKEY_DECREASE:
#else
    case '-':
#endif
      map_set_zoom(appdata->map, 
		   appdata->map->zoom/ZOOM_FACTOR_BUTTON);
      printf("zoom is now %f\n", appdata->map->zoom);
      return TRUE;
      break;

    default:
      printf("key event %d\n", event->keyval);
      break;
    }
  }

  return FALSE;
}

GtkWidget *map_new(appdata_t *appdata) {
  map_t *map = appdata->map = g_new0(map_t, 1);

  map->zoom = 0.25;
  map->pen_down.at.x = -1;
  map->pen_down.at.y = -1;
  map->appdata = appdata;
  map->action.type = MAP_ACTION_IDLE;

#ifdef USE_GNOMECANVAS
  map->canvas = gnome_canvas_new_aa();  // _aa

  /* create the groups */
  canvas_group_t group;
  for(group = 0; group < CANVAS_GROUPS; group++) 
    map->group[group] = gnome_canvas_item_new(
	       	gnome_canvas_root(GNOME_CANVAS(map->canvas)),
	       	GNOME_TYPE_CANVAS_GROUP,
	       	NULL);       

  gtk_widget_modify_bg(map->canvas, GTK_STATE_NORMAL, 
		       &map->canvas->style->white);

#else
  map->canvas = goo_canvas_new();

  g_object_set(G_OBJECT(map->canvas), "anchor", GTK_ANCHOR_CENTER, NULL);

  GooCanvasItem *root = goo_canvas_get_root_item(GOO_CANVAS(map->canvas));

  /* create the groups */
  canvas_group_t group;
  for(group = 0; group < CANVAS_GROUPS; group++) 
    map->group[group] = goo_canvas_group_new(root, NULL);       

#endif

  gtk_widget_set_events(map->canvas,
                          GDK_BUTTON_PRESS_MASK
			| GDK_BUTTON_RELEASE_MASK
			| GDK_SCROLL_MASK
			| GDK_POINTER_MOTION_MASK
    			| GDK_POINTER_MOTION_HINT_MASK);
  
  gtk_signal_connect(GTK_OBJECT(map->canvas), 
     "button_press_event", G_CALLBACK(map_button_event), appdata);
  gtk_signal_connect(GTK_OBJECT(map->canvas), 
     "button_release_event", G_CALLBACK(map_button_event), appdata);
  gtk_signal_connect(GTK_OBJECT(map->canvas), 
     "motion_notify_event", G_CALLBACK(map_motion_notify_event), appdata);
  gtk_signal_connect(GTK_OBJECT(map->canvas), 
     "scroll_event", G_CALLBACK(map_scroll_event), appdata);

  gtk_signal_connect(GTK_OBJECT(map->canvas), 
     "destroy", G_CALLBACK(map_destroy_event), appdata);

  return map->canvas;
}

void map_init(appdata_t *appdata) {
  map_t *map = appdata->map;

  /* set initial zoom */
  map_set_zoom(map, map->zoom);
  map_init_colors(appdata->potlatch, appdata->osm);
  map_draw(map, appdata->osm);

  canvas_set_bounds(map->canvas,
	appdata->osm->bounds->min.x, appdata->osm->bounds->min.y,
	appdata->osm->bounds->max.x, appdata->osm->bounds->max.y);

  canvas_scroll_to(map->canvas, 0,0);
}


void map_item_set_flags(map_item_t *map_item, int set, int clr) {

  switch(map_item->type) {
  case MAP_TYPE_NODE:
    map_item->node->flags |=  set;
    map_item->node->flags &= ~clr;
    break;

  case MAP_TYPE_WAY:
    map_item->way->flags |=  set;
    map_item->way->flags &= ~clr;
    break;

  default:
    g_assert(0);
    break;
  }
}

void map_clear(appdata_t *appdata) {
  map_t *map = appdata->map;

  printf("freeing map contents\n");

  /* remove a possibly existing highlight */
  map_item_deselect(appdata);
  
  canvas_group_t group;
  for(group=0;group<CANVAS_GROUPS;group++) {

    /* destroy the entire group */
    canvas_item_destroy(map->group[group]);

#ifdef USE_GNOMECANVAS
    /* and create an empty new one */
    map->group[group] = gnome_canvas_item_new(
		      gnome_canvas_root(GNOME_CANVAS(map->canvas)),
		      GNOME_TYPE_CANVAS_GROUP,
		      NULL);       
#else
    GooCanvasItem *root = goo_canvas_get_root_item(GOO_CANVAS(map->canvas));
    map->group[group] = goo_canvas_group_new(root, NULL);       
#endif
  }
}

void map_paint(appdata_t *appdata) {
  map_t *map = appdata->map;

  map_init_colors(appdata->potlatch, appdata->osm);
  map_draw(map, appdata->osm);
}

/* called from several icons like e.g. "node_add" */
void map_action_set(appdata_t *appdata, map_action_t action) {
  printf("map action set to %d\n", action);

  appdata->map->action.type = action;

  /* enable/disable ok/cancel buttons */
  // MAP_ACTION_IDLE=0, NODE_ADD, BG_ADJUST, WAY_ADD, WAY_NODE_ADD, WAY_CUT
  const gboolean ok_state[] = { FALSE, FALSE, TRUE, FALSE, FALSE, FALSE };
  const gboolean cancel_state[] = { FALSE, TRUE, TRUE, TRUE, TRUE, TRUE };

  g_assert(MAP_ACTION_NUM == sizeof(ok_state)/sizeof(gboolean));
  g_assert(action < sizeof(ok_state)/sizeof(gboolean));

  icon_bar_map_cancel_ok(appdata, cancel_state[action], ok_state[action]);

  switch(action) {
  case MAP_ACTION_BG_ADJUST:
    /* an existing selection only causes confusion ... */
    map_item_deselect(appdata);
    break;

  case MAP_ACTION_WAY_ADD:
    printf("starting new way\n");
    map_item_deselect(appdata);
    g_assert(!appdata->map->action.way);
    appdata->map->action.way = osm_way_new();
    break;

  case MAP_ACTION_NODE_ADD:
    map_item_deselect(appdata);
    break;

  default:
    break;
  }

  icon_bar_map_action_idle(appdata, action == MAP_ACTION_IDLE);
  gtk_widget_set_sensitive(appdata->menu_item_wms_adjust, 
			   action == MAP_ACTION_IDLE);

  const char *str_state[] = { 
    NULL, 
    _("Place a node"),
    _("Adjust background image position"),
    _("Place first node of new way"),
    _("Place node on selected way"),
    _("Select segment to cut way"),
  };

  g_assert(MAP_ACTION_NUM == sizeof(str_state)/sizeof(char*));

  statusbar_set(appdata, str_state[action]);
}


void map_action_cancel(appdata_t *appdata) {
  map_t *map = appdata->map;

  switch(map->action.type) {
  case MAP_ACTION_WAY_ADD:
    printf("  removing temporary way\n");
    g_assert(map->action.way);

    /* remove all nodes that have been created for this way */
    /* (their way count will be 0 after removing the way) */
    node_chain_t *chain = map->action.way->node_chain;
    while(chain) {
      node_chain_t *next = chain->next;

      printf("    node #%ld (used by %d)\n", 
	     chain->node->id, chain->node->ways);

      chain->node->ways--;
      if(!chain->node->ways && (chain->node->id == ID_ILLEGAL)) {
	printf("      -> freeing temp node\n");
	osm_node_free(chain->node);
      }
      g_free(chain);
      chain = next;
    }
    map->action.way->node_chain = NULL;

    /* remove ways visual representation */
    map_item_chain_destroy(&map->action.way->map_item_chain);

    osm_way_free(map->action.way);
    map->action.way = NULL;
    break;

  case MAP_ACTION_BG_ADJUST:
    /* undo all changes to bg_offset */
    map->bg_offset.x = appdata->project->wms_offset.x;
    map->bg_offset.y = appdata->project->wms_offset.y;

    gint x = appdata->osm->bounds->min.x + map->bg_offset.x;
    gint y = appdata->osm->bounds->min.y + map->bg_offset.y;
    canvas_image_move(map->bg_item, x, y, map->bg_scale.x, map->bg_scale.y);
    break;

  default:
    break;
  }

  map_action_set(appdata, MAP_ACTION_IDLE);
}

void map_action_ok(appdata_t *appdata) {
  map_t *map = appdata->map;

  switch(map->action.type) {
  case MAP_ACTION_WAY_ADD:
    g_assert(map->action.way);
    
    /* transfer all nodes that have been created for this way */
    /* into the node chain */

    /* (their way count will be 0 after removing the way) */
    node_chain_t *chain = map->action.way->node_chain;
    while(chain) {
      printf("    node #%ld (used by %d)\n", 
	     chain->node->id, chain->node->ways);

      /* a node may have been a stand-alone node before, so remove its */
      /* visual representation as its now drawn as part of the way */
      /* (if at all) */
      if(chain->node->id != ID_ILLEGAL) 
	map_item_chain_destroy(&chain->node->map_item_chain);

      /* we can be sure that no node gets inserted twice (even if twice in */
      /* the ways chain) because it gets assigned a non-ID_ILLEGAL id when */
      /* being moved to the osm node chain */
      if(chain->node->id == ID_ILLEGAL) 
	osm_node_attach(appdata->osm, chain->node);

      chain = chain->next;
    }

    /* now move the way itself into the main data structure */
    osm_way_attach(appdata->osm, map->action.way);

    map_way_select(appdata, map->action.way);

    map->action.way = NULL;

    /* let the user specify some tags for the new way */
    info_dialog(appdata);

    break;

  case MAP_ACTION_BG_ADJUST:
    /* save changes to bg_offset in project */
    appdata->project->wms_offset.x = map->bg_offset.x;
    appdata->project->wms_offset.y = map->bg_offset.y;
    appdata->project->dirty = TRUE;
    break;

  default:
    break;
  }

  map_action_set(appdata, MAP_ACTION_IDLE);
}

/* called from icon "trash" */
void map_delete_selected(appdata_t *appdata) {
  map_t *map = appdata->map;

  /* work on local copy since de-selecting destroys the selection */
  map_item_t item = map->selected;

  /* deleting the selected item de-selects it ... */
  map_item_deselect(appdata);

  switch(item.type) {
  case MAP_TYPE_NODE:
    printf("request to delete node #%ld\n", item.node->id);

    /* check if this node is part of a way with two nodes only. */
    /* we cannot delete this as this would also delete the way */
    way_chain_t *way_chain = osm_node_to_way(appdata->osm, item.node);
    if(way_chain) {
      gboolean short_way = FALSE;
      
      /* free the chain of ways */
      while(way_chain && !short_way) {
	way_chain_t *next = way_chain->next;

	if(osm_way_number_of_nodes(way_chain->way) <= 2) 
	  short_way = TRUE;

	g_free(way_chain);
	way_chain = next;
      }

      if(short_way) {
	if(!yes_no_f(GTK_WIDGET(appdata->window), 
		     _("Delete node in short way(s)?"),
		     _("Deleting this node will also delete one or more ways "
		       "since they'll contain only one node afterwards. "
		       "Do you really want this?")))
	  return;
      }
    }

    /* remove it visually from the screen */
    map_item_chain_destroy(&item.node->map_item_chain);

    /* and mark it "deleted" in the database */
    osm_node_remove_from_relation(appdata->osm, item.node);
    way_chain_t *chain = osm_node_delete(appdata->osm, item.node, FALSE, TRUE);

    /* redraw all affected ways */
    while(chain) {
      way_chain_t *next = chain->next;

      if(osm_way_number_of_nodes(chain->way) == 1) {
	/* this way now only contains one node and thus isn't a valid */
	/* way anymore. So it'll also get deleted (which in turn may */
	/* cause other nodes to be deleted as well) */
	map_way_delete(appdata, chain->way);
      } else {
	map_item_t item;
	item.type = MAP_TYPE_WAY;
	item.way = chain->way;
	map_item_redraw(appdata, &item);
      }

      g_free(chain);

      chain = next;
    }

    break;

  case MAP_TYPE_WAY:
    printf("request to delete way #%ld\n", item.way->id);
    map_way_delete(appdata, item.way);
    break;

  default:
    g_assert((item.type == MAP_TYPE_NODE) ||
	     (item.type == MAP_TYPE_WAY));
    break;
  }
}

/* ----------------------- track related stuff ----------------------- */

void map_track_draw_seg(map_t *map, track_seg_t *seg) {
  /* a track_seg needs at least 2 points to be drawn */
  guint pnum = track_seg_points(seg);
  printf("seg of length %d\n", pnum);

  if(pnum == 1) {
    g_assert(!seg->item);

    seg->item = canvas_circle_new(map, CANVAS_GROUP_TRACK,
		  seg->track_point->lpos.x, seg->track_point->lpos.y, 
		  1.5*WAY_WIDTH, 0, TRACK_COLOR, NO_COLOR);
  }

  if(pnum > 1) {
    
    /* allocate space for nodes */
    canvas_points_t *points = canvas_points_new(pnum);
    
    int point = 0;
    track_point_t *track_point = seg->track_point;
    while(track_point) {
      points->coords[point++] = track_point->lpos.x;
      points->coords[point++] = track_point->lpos.y;
      track_point = track_point->next;
    }
    
    /* there may be a circle (one point line) */
    if(seg->item)
      canvas_item_destroy(seg->item);

    seg->item = canvas_polyline_new(map, CANVAS_GROUP_TRACK,
		  points, 3*WAY_WIDTH, TRACK_COLOR);

    canvas_points_free(points);
  }
}

void map_track_update_seg(map_t *map, track_seg_t *seg) {
  /* a track_seg needs at least 2 points to be drawn */
  guint pnum = track_seg_points(seg);
  printf("seg of length %d\n", pnum);

  if(pnum > 1) {
    
    /* allocate space for nodes */
    canvas_points_t *points = canvas_points_new(pnum);
    
    int point = 0;
    track_point_t *track_point = seg->track_point;
    while(track_point) {
      canvas_point_set_pos(points, point++, &track_point->lpos);
      track_point = track_point->next;
    }
    
    g_assert(seg->item);
    canvas_item_set_points(seg->item, points);
    canvas_points_free(points);
  }
}

void map_track_draw(map_t *map, track_t *track) {
  track_seg_t *seg = track->track_seg;

  /* draw all segments */
  while(seg) {
    map_track_draw_seg(map, seg);
    seg = seg->next;
  }
}

void map_track_remove(appdata_t *appdata) {
  track_t *track = appdata->track.track;

  printf("removing track\n");

  g_assert(track);

  /* remove all segments */
  track_seg_t *seg = track->track_seg;
  while(seg) {
    if(seg->item) {
      canvas_item_destroy(seg->item);
      seg->item = NULL;
    }

    seg = seg->next;
  }
}

void map_track_pos(appdata_t *appdata, lpos_t *lpos) {
  if(appdata->track.gps_item) {
    canvas_item_destroy(appdata->track.gps_item);
    appdata->track.gps_item = NULL;
  }

  if(lpos)
    appdata->track.gps_item = canvas_circle_new(appdata->map, CANVAS_GROUP_GPS, 
	lpos->x, lpos->y, 1.5 * WAY_WIDTH, 0, TRACK_GPS_COLOR, NO_COLOR);
}

/* ------------------- map background ------------------ */

void map_remove_bg_image(map_t *map) {
  if(!map) return;

  if(map->bg_item) {
    canvas_item_destroy(map->bg_item);
    map->bg_item = NULL;
  }
}

static gint map_bg_item_destroy_event(GtkWidget *widget, gpointer data) {
  map_t *map = (map_t*)data;

  /* destroying background item */

  map->bg_item = NULL;
  if(map->bg_pix) {
    printf("destroying background item\n");
    gdk_pixbuf_unref(map->bg_pix);
    map->bg_pix = NULL;
  }
  return FALSE;
}

void map_set_bg_image(map_t *map, char *filename) {
  bounds_t *bounds = map->appdata->osm->bounds;

  map_remove_bg_image(map);

  map->bg_pix = gdk_pixbuf_new_from_file(filename, NULL);

  /* calculate required scale factor */
  map->bg_scale.x = (float)(bounds->max.x - bounds->min.x)/
    (float)gdk_pixbuf_get_width(map->bg_pix);
  map->bg_scale.y = (float)(bounds->max.y - bounds->min.y)/
    (float)gdk_pixbuf_get_height(map->bg_pix);

  map->bg_item = canvas_image_new(map, CANVAS_GROUP_BG, map->bg_pix, 
	  bounds->min.x, bounds->min.y, map->bg_scale.x, map->bg_scale.y);

  canvas_item_destroy_connect(map->bg_item, 
          G_CALLBACK(map_bg_item_destroy_event), map);
}  
